from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector

cdef extern from "jly_goicp.hpp":
    cdef cppclass POINT3D:
        float x
        float y
        float z
        POINT3D()
        POINT3D(float, float, float)
        void pointToString()
    
    cdef cppclass ROTNODE:
        float a
        float b
        float c
        float w
        float ub
        float lb
        int l
        ROTNODE()
    
    cdef cppclass TRANSNODE:
        float x
        float y
        float z 
        float w
        float ub
        float lb
        TRANSNODE()
        
    cdef cppclass GoICP:
        int Nm
        int Nd
        float MSEThresh
        float SSEThresh
        float icpThresh
        float trimFraction
        int inlierNum
        bool doTrim
    
        GoICP()
        float Register() except +
        void BuildDT() except + 
        libcpp_vector[libcpp_vector[double]] optimalRotation() except +
        libcpp_vector[double] optimalTranslation() except +
        
        void loadModelAndData(int, libcpp_vector[POINT3D], int, libcpp_vector[POINT3D]) except +
        void setInitNodeRot(ROTNODE&) except +
        void setInitNodeTrans(TRANSNODE&) except +
        void setDTSizeAndFactor(int, double) except +


import os, platform, pkg_resources
# from distutils.core import setup, Extension
from setuptools import setup, Extension;
from Cython.Distutils import build_ext

NO_AUTOWRAP=False;
try:
    import autowrap;
except ImportError:
    NO_AUTOWRAP=True;

VERSION = (0, 0, 4);

# with open("README.md", "r") as fh:
#     long_description = fh.read()

long_description ="Go-ICP for globally optimal 3D pointset registration.";

data_dir = pkg_resources.resource_filename("autowrap", "data_files")
include_dir = os.path.join(data_dir, "autowrap")

if(NO_AUTOWRAP):
    raise IndentationError;
else:
    print('FOUND AUTOWRAP LIBRARY, PROCEED');

print('PRECOMPILE USING AUTOWRAP BEFORE PROCEEDING');
from autowrap.Main import run as autowrap_run;
autowrap_run([os.path.abspath('./src/goicpcc.pxd')], [], [], os.path.abspath('./src/py_goicp.pyx'), );

# if(not os.path.exists(os.path.abspath('./py_chenhancc.cpp'))):
#     import subprocess;
#     source_files_dir = os.path.abspath("./src");
#     source_file_pxd = os.path.join(source_files_dir, "chenhancc.pxd");
#     out_file_pyx = os.path.join(source_files_dir, "py_chenhancc.pyx");
#     
#     subprocess.Popen(" ".join(["autowrap", "--out", out_file_pyx, source_file_pxd]), cwd=os.path.abspath("./src"));
# #     autowrap --out py_chenhancc.pyx chenhancc.pxd

ext = Extension("py_goicp",
                sources = ['src/py_goicp.cpp'],
                language="c++",
                extra_compile_args=["-std=c++14"], #Release mode (no -g switch)
                extra_link_args=["-std=c++14"], #Release mode (no -g switch)
#                 extra_compile_args=["-std=c++14", "-g"], #Debug mode (no -g switch)
#                 extra_link_args=["-std=c++14",  "-g"], #Debug mode (no -g switch)
                include_dirs = [include_dir, data_dir],
               )

setup(cmdclass={'build_ext':build_ext},
      name="py_goicp",
      version="%d.%d.%d" % VERSION,
      ext_modules = [ext],
      install_requires=['autowrap>=0.22'],
      author='#0K Srinivasan Ramachandran',
      author_email='ashok.srinivasan2002@gmail.com',
      url='https://github.com/aalavandhaann/go-icp_cython',
      maintainer="#0K Srinivasan Ramachandran",
      maintainer_email="ashok.srinivasan2002@gmail.com",
      platforms=["any"],
      description='GO-ICP compiled using Cython to use in python',
      license='LICENSE.txt',
      keywords='icp go-icp registration alignment rigid-align rigid-alignment',
      python_requires='>=2',
      long_description_content_type='text/markdown',
      long_description=open('README.md').read(),
      zip_safe=False,
     )

###AUTOWRAP
#autowrap --out py_chenhan.pyx chenhancc.pxd
#python setup.py build_ext --inplace


from py_goicp import GoICP, POINT3D, ROTNODE, TRANSNODE;
import numpy as np;
import time;

def loadPointCloud(filename):
    pcloud = np.loadtxt(filename, skiprows=1);
    plist = pcloud.tolist();
    p3dlist = [];
    for x,y,z in plist:
        pt = POINT3D(x,y,z);
        p3dlist.append(pt);
    return pcloud.shape[0], p3dlist, pcloud;

goicp = GoICP();
rNode = ROTNODE();
tNode = TRANSNODE();
 
rNode.a = -3.1416;
rNode.b = -3.1416;
rNode.c = -3.1416;
rNode.w = 6.2832;
 
tNode.x = -0.5;
tNode.y = -0.5;
tNode.z = -0.5;
tNode.w = 1.0;

goicp.MSEThresh = 0.001;
goicp.trimFraction = 0.0;
  
if(goicp.trimFraction < 0.001):
    goicp.doTrim = False;

a_points = [POINT3D(0.0, 0.0, 0.0), POINT3D(0.5, 1.0, 0.0), POINT3D(1.0, 0.0, 0.0)];
b_points = [POINT3D(0.0, 0.0, 0.0), POINT3D(1.0, 0.5, 0.0), POINT3D(1.0, -0.5, 0.0)];

Nm, a_points, np_a_points = loadPointCloud('./tests/model_bunny.txt');
Nd, b_points, np_b_points = loadPointCloud('./tests/data_bunny.txt');

goicp.loadModelAndData(Nm, a_points, Nd, b_points);
#LESS DT Size = LESS TIME CONSUMPTION = HIGHER ERROR
goicp.setDTSizeAndFactor(300, 2.0);
goicp.setInitNodeRot(rNode);
goicp.setInitNodeTrans(tNode);

start = time.time();
print("Building Distance Transform...");
goicp.BuildDT();
print("REGISTERING....");
goicp.Register();
end = time.time();
print('TOTAL TIME : ', (end-start));
optR = np.array(goicp.optimalRotation());
optT = goicp.optimalTranslation();
optT.append(1.0);
optT = np.array(optT);

transform = np.empty((4,4));
transform[:3, :3] = optR;
transform[:,3] = optT;

print(np_b_points.shape, np.ones((Nd, 1)).shape);

#Now transform the data mesh to fit the model mesh
transform_model_points = (transform.dot(np.hstack((np_b_points, np.ones((Nd, 1)))).T)).T;
transform_model_points = transform_model_points[:,:3];

PLY_FILE_HEADER = "ply\nformat ascii 1.0\ncomment PYTHON generated\nelement vertex %s\nproperty float x\nproperty float y\nproperty float z\nend_header"%(Nd);

np.savetxt('./tests/data_bunny_transformed.ply', transform_model_points, header = PLY_FILE_HEADER, comments='');

print(optR);
print(optT);
print(transform);




